// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "StateEvent.hpp"
#include "StateProperty.hpp"

namespace Dawn {
  class StateOwner {
    private:
      std::vector<IStateEvent*> eventsSubscribed;

    public:
      /**
       * Listen for changees to a state property and involke the provided func
       * when the value is changed.
       * 
       * @tparam V The type of the state property.
       * @param property Property to listen for affect changees to.
       * @param fn The callback to be invoked when the state value changes.
       */
      template<class V>
      void useEffect(
        StateProperty<V> &property,
        const std::function<void()> &fn
      ) {
        property._effectListners.push_back(fn);
      }

      /**
       * Listen for when an event is invoked by a state event. This is intended
       * to allow for cross-state-owner communication in a simple and effective
       * way.
       * 
       * @tparam F The type of the callback function.
       * @tparam A The arguments from the state event that are calledback.
       * @param event The event that is being subscribed to.
       * @param fn The function to be inokved on event trigger.
       */
      template<typename F, typename... A>
      void useEvent(StateEvent<A...> &event, F fn) {
        event._eventListeners[this].push_back(fn);
        this->eventsSubscribed.push_back(&event); 
      }

      ~StateOwner() {
        auto it = this->eventsSubscribed.begin();
        while(it != this->eventsSubscribed.end()) {
          (*it)->_stateOwnerDestroyed(this);
          ++it;
        }
      }
  };
}